#pragma once

#include <vector>
#include <memory>

class MarkdownElement
{
public:
    enum class Type
    {
        HEADING,
        PARAGRAPH,
        TEXT_SPAN,
        INLINE_CODE,
        MULTILINE_CODE,
        INLINE_QUOTE,
        MULTILINE_QUOTE,
        INLINE_SPECIAL,
        MULTILINE_SPECIAL,
        LINK,
        IMAGE
    };

    virtual ~MarkdownElement() = default;

    void appendTextContent(const std::string& content)
    {
        mTextContent += content;
    }

    const std::string& getTextContent() const
    {
        return mTextContent;
    }

    virtual Type getType() const = 0;
private:
    std::string mTextContent;
};

class MarkdownInlineElement : public MarkdownElement
{
public:
    virtual ~MarkdownInlineElement() = default;
};

class MarkdownTextSpan : public MarkdownInlineElement
{
public:
    virtual ~MarkdownTextSpan() = default;

    Type getType() const override
    {
        return Type::TEXT_SPAN;
    }
};

class MarkdownParagraph : public MarkdownElement
{
public:
    virtual ~MarkdownParagraph() = default;

    Type getType() const override
    {
        return Type::PARAGRAPH;
    }

    void addChild(std::unique_ptr<MarkdownInlineElement> child)
    {
        mChildren.push_back(std::move(child));
    }

    std::size_t getNumChildren() const
    {
        return mChildren.size();
    }

    MarkdownInlineElement* getChild(std::size_t idx) const
    {
        return mChildren[idx].get();
    }

    std::vector<std::unique_ptr<MarkdownInlineElement> > mChildren;
};

class MarkdownHeading : public MarkdownElement
{
public:
    MarkdownHeading(unsigned level)
        : mLevel(level)
    {

    }
    virtual ~MarkdownHeading() = default;

    Type getType() const override
    {
        return Type::HEADING;
    }

    unsigned getLevel() const
    {
        return mLevel;
    }

private:
    unsigned mLevel{1};
};



class MarkdownInlineQuote : public MarkdownInlineElement
{
public:

    virtual ~MarkdownInlineQuote() = default;

    Type getType() const override
    {
        return Type::INLINE_QUOTE;
    }
};

class MarkdownMultilineQuote : public MarkdownElement
{
public:
    MarkdownMultilineQuote(const std::string& tag)
        : mTag(tag)
    {

    }

    virtual ~MarkdownMultilineQuote() = default;

    Type getType() const override
    {
        return Type::MULTILINE_QUOTE;
    }

private:
    std::string mTag;
};


class MarkdownDocument
{
public:
    void addElement(std::unique_ptr<MarkdownElement> element)
    {
        mElements.push_back(std::move(element));
    }

    std::size_t getNumElements() const
    {
        return mElements.size();
    }

    MarkdownElement* getElement(std::size_t idx) const
    {
        return mElements[idx].get();
    }

private:
    std::vector<std::unique_ptr<MarkdownElement> > mElements;
};


